import unittest
import numpy
import PandaSurvey as ps
from PandaSurvey.simple import SimpleRake


class TestRake(unittest.TestCase):

    def setUp(self):
        self.recodes = {
            "Age": lambda age: int(age / 10),
            "MaritalStatus": lambda m: int(m / 2)
        }

        self.target_proportions = {
            "Gender": {
                1: .4916,
                2: .5084
            }
        }

        self.df = ps.load_people()

    def test_smoke(self):
        r = SimpleRake(self.df, self.target_proportions)
        self.assertIsNotNone(r)

    def test_recode(self):
        r = SimpleRake(self.df, self.target_proportions)
        r.recode(self.recodes)
        self.assertEqual(r.df['Age'].values[0], 8)
        self.assertEqual(r.df['MaritalStatus'].values[2], 2)


class TestSimpleRake(unittest.TestCase):

    def setUp(self):
        self.prop = ps.load_sample_proportions()
        self.df = ps.load_sample_study()
        self.wt = ps.load_sample_weights()
        self.rake = SimpleRake(self.df, self.prop)

    def calc_mse(self, df):
        results = []
        alt = []
        targets = []
        wt_sum = df['weight'].sum()
        alt_wt_sum = df['target_weight'].sum()
        for demo in self.prop:
            for value in self.prop[demo]:
                subset = df[self.df[demo] == int(value)]
                results.append(subset['weight'].sum() / wt_sum)
                alt.append(subset['target_weight'].sum() / alt_wt_sum)
                targets.append(self.prop[demo][value])

        targets = numpy.array(targets)
        results = numpy.array(results)
        alt = numpy.array(alt)

        mse = ((results - targets) ** 2).mean()
        altmse = ((alt - targets) ** 2).mean()
        return mse, altmse

    def test_calc_l1(self):
        df_out = self.rake.calc()
        df_out['target_weight'] = self.wt
        mse, altmse = self.calc_mse(df_out)
        self.assertLessEqual(mse, altmse)
        self.assertLessEqual(self.rake.loss(df_out.weight), 1e-3)

    def test_calc_l2(self):
        df_out = self.rake.calc(use_l2=True)
        df_out['target_weight'] = self.wt
        mse, altmse = self.calc_mse(df_out)
        self.assertLessEqual(mse, altmse)
        self.assertLessEqual(self.rake.loss(df_out.weight), 1e-3)



"""PandaSurvey includes two unique datasets for testing purpuses: `People` and a sample study. The `People` file is from the 2010 US Census. The sample study is from a small survey performed at InContext Solutions in 2014 (specific survey details withheld)"""

import os
import pandas


def _path(name):
    root, _ = os.path.split(__file__)
    return os.path.join(root, 'data/' + name)


def load_people():
    """Returns the `People` dataset as a DataFrame. The data consists of 9999 individuals with age, disability status, marital status, race, and gender demographic information. Columns and their codes are described below:

    - Age
        - Non-negative integer
        - May include zeros
    - Disability
        - 1: Disabled
        - 2: Not disabled
    - MarritalStatus
        - 1: Married
        - 2: Widowed
        - 3: Divorced
        - 4: Separated
        - 5: Never married or under 15 years old
    - Race
        - 1: White alone
        - 2: Black or African American alone
        - 3: American Indian alone
        - 4: Alaska Native alone
        - 5: American Indian and Alaska Native tribes specified; or American Indian or Alaska native, not specified and no other races
        - 6: Asian alone
        - 7: Native Hawaiian and Other Pacific Islander alone
        - 8: Some other race alone
        - 9: Two or more major race groups
    - Gender
        - 1: Male
        - 2: Female
    """
    return pandas.read_csv(_path("People.csv"))


def load_sample_study():
    """Returns a sample dataset describing demographics in coded format from 2092 respondents. The study consists of 7 cells and demographics considered include age, gender, income, hispanic, and race."""
    df = pandas.read_csv(_path("SampleStudy.csv"))
    del df['Weight']
    return df


def load_sample_weights():
    """Returns individual weights from the sample survey calculated via a raking method previously implemented in R."""
    df = pandas.read_csv(_path("SampleStudy.csv"))
    return df['Weight']


def load_sample_proportions():
    """Returns the target sample proportions that correspond to the sample survey.

    +-------------+-------------+-------------------+
    | Demographic | Coded Value | Target Proportion |
    +=============+=============+===================+
    | Age         | 1           | 0.07              |
    +-------------+-------------+-------------------+
    | Age         | 2           | 0.22              |
    +-------------+-------------+-------------------+
    | Age         | 3           | 0.2               |
    +-------------+-------------+-------------------+
    | Age         | 4           | 0.2               |
    +-------------+-------------+-------------------+
    | Age         | 5           | 0.21              |
    +-------------+-------------+-------------------+
    | Gender      | 1           | 0.5               |
    +-------------+-------------+-------------------+
    | Gender      | 2           | 0.5               |
    +-------------+-------------+-------------------+
    | Income      | 1           | 0.17              |
    +-------------+-------------+-------------------+
    | Income      | 2           | 0.21              |
    +-------------+-------------+-------------------+
    | Income      | 3           | 0.25              |
    +-------------+-------------+-------------------+
    | Income      | 4           | 0.16              |
    +-------------+-------------+-------------------+
    | Income      | 5           | 0.11              |
    +-------------+-------------+-------------------+
    | Hispanic    | 1           | 0.09              |
    +-------------+-------------+-------------------+
    | Hispanic    | 2           | 0.91              |
    +-------------+-------------+-------------------+
    | Race        | 0           | 0.15              |
    +-------------+-------------+-------------------+
    | Race        | 1           | 0.85              |
    +-------------+-------------+-------------------+
    """
    weights = {}
    with open(_path("SampleWeights.csv")) as csv_in:
        for line in csv_in:
            demo, category, proportion = line.split(',')
            if demo not in weights:
                weights[demo] = {}
            weights[demo][int(category)] = float(proportion)
    return weights



import unittest
import PandaSurvey as ps


class TestLoad(unittest.TestCase):

    def test_load_people(self):
        df = ps.load_people()
        self.assertEqual(df.shape, (9999, 5))

    def test_load_sample_proportions(self):
        targets = ps.load_sample_proportions()
        self.assertEqual(type(targets), type({}))
        self.assertEqual(len(targets), 5)
        self.assertEqual(len(targets['Age']), 5)

    def test_load_sample_weights(self):
        df = ps.load_sample_weights()
        self.assertEqual(df.shape, (2092,))

    def test_load_sample_study(self):
        df = ps.load_sample_study()
        self.assertEqual(df.shape, (2092, 6))



import numpy
import pandas
from .base import SurveyWeightBase


BIG_M = 99
EPSILON = 1e-3
MAX_ITER = 1000


class SimpleRake(SurveyWeightBase):

    """Rake weighting implementation.

    :param pandas.DataFrame df: The observations to be weighted.
    :param dict proportions: A dictionary of the target proportions for each demographic and response.
    :param float epsilon: Convergence threshold for the raking procedure.
    :param int maxiter: Maximum number of iterations for the raking procedure.
    """

    def __init__(self, df, proportions, epsilon=EPSILON, maxiter=MAX_ITER):
        self.df = df
        self.rows, self.cols = self.df.shape
        self.proportions = proportions
        self.demos = self.proportions.keys()
        self.epsilon = epsilon
        self.maxiter = maxiter

    def calc(self, use_l2=False):
        """Calculates individual weights.

        :param boolean use_l2: Determines if convergence is measured using the L2 norm of the change in weights. By default, the L1 norm is used.
        """
        warning_state = pandas.options.mode.chained_assignment
        pandas.options.mode.chained_assignment = None

        delta = BIG_M
        delta0 = BIG_M + 1
        num_iters = 0

        new_df = self.df.copy()
        new_df['weight'] = numpy.ones(self.rows)

        def _update(row, mass):
            value = int(row[var])
            return self.proportions[var][value] * row['weight'] / (mass[value] / self.rows)

        try:
            while delta < delta0 * (1 - self.epsilon) and num_iters < self.maxiter:
                wt = new_df['weight']

                for var in self.proportions:
                    mass = {group[0]: group[1].sum()['weight'] for group in new_df.groupby(var)}
                    sub_df = new_df[new_df[var].isin(self.proportions[var])]
                    sub_df['weight'] = sub_df.apply(_update, axis=1, mass=mass)
                    new_df.update(sub_df)

                delta0 = delta

                if use_l2:
                    delta = ((new_df['weight'] - wt) ** 2).sum()
                else:
                    delta = (new_df['weight'] - wt).abs().sum()

                num_iters += 1
        finally:
            pandas.options.mode.chained_assignment = warning_state

        new_df[self.demos] = new_df[self.demos].astype(int)
        return new_df



from abc import ABCMeta, abstractmethod


class SurveyWeightBase:

    """Abstract base class for survey weighting method."""

    __metaclass__ = ABCMeta

    def __init__(self, df, proportions, recodes={}):
        pass

    @abstractmethod
    def calc(self):
        """Calculation step must be implemented here."""
        pass

    def loss(self, weights):
        """Describes the inflation in the variance of sample estimates that can be attributed to weighting. See *Applied Survey Data Analysis* (2010) by Heeringa et al. for more information.

        :param numpy.array weights: array of individual weights
        """
        n = len(weights)
        return ((weights ** 2).sum() / weights.sum() ** 2) * n - 1

    def recode(self, encoders):
        """Recodes demographic information.

        :param dict encoders: Mapping of demographic keys to respective recoding function. Need only define those demographic dimensions that need to be recoded.
        """
        for demo in encoders:
            func = lambda row: encoders[demo](row[demo])
            self.df[demo] = self.df.apply(func, axis=1)



from setuptools import setup, find_packages


install_requires = [
    'numpy',
    'pandas',
]

tests_require = [
    'nose',
    'coverage'
]

setup(
    name="PandaSurvey",
    version="0.111",
    author="InContext Solutions",
    author_email="quant@incontextsolutions.com",
    url="http://www.incontextsolutions.com/",
    download_url="https://github.com/InContextSolutions/PandaSurvey/tarball/v0.1",
    description="Survey weighting utility for the Pandas DataFrame",
    keywords=['raking', 'Pandas', 'survey'],
    packages=find_packages(),
    setup_requires=['nose'],
    install_requires=install_requires,
    tests_require=tests_require,
    entry_points={},
    classifiers=[]
)
